# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._base.extensions.entity import Extension


def build_extension(extensions: list[dict] | None = None) -> list[Extension]:
    """
    Build a list of Extension objects from a list of dictionaries.

    Parameters
    ----------
    extensions : list[dict]
        List of extension dictionaries.

    Returns
    -------
    list[Extension]
        List of Extension objects.
    """
    if extensions is None:
        return []
    return [Extension(**ext_dict) for ext_dict in extensions]



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing
from concurrent.futures import ThreadPoolExecutor

from digitalhub.entities._base.executable.entity import ExecutableEntity
from digitalhub.entities._commons.enums import EntityTypes, Relationship
from digitalhub.factory.entity import entity_factory

if typing.TYPE_CHECKING:
    from digitalhub.entities.function._base.spec import FunctionSpec
    from digitalhub.entities.function._base.status import FunctionStatus
    from digitalhub.entities.run._base.entity import Run
    from digitalhub.entities.task._base.entity import Task


class Function(ExecutableEntity):
    """
    A class representing a function.
    """

    ENTITY_TYPE = EntityTypes.FUNCTION.value

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.spec: FunctionSpec
        self.status: FunctionStatus

    ##############################
    #  Function Methods
    ##############################

    def run(
        self,
        action: str,
        local_execution: bool = False,
        wait: bool = False,
        log_info: bool = True,
        **kwargs,
    ) -> Run:
        """
        Run function. This method creates a new run and executes it.

        Parameters
        ----------
        action : str
            Action to execute.
        local_execution : bool
            Flag to determine if object has local execution.
        wait : bool
            Flag to wait for execution.
        log_info : bool
            Flag to log information while waiting.
        **kwargs : dict
            Keyword arguments passed to Run builder.

        Returns
        -------
        Run
            Run instance.
        """
        # Create or update new task
        task: Task = self._get_or_create_task(action)

        # Run function from task
        run_kind = entity_factory.get_run_kind_from_action(self.kind, action)
        run = task.run(run_kind, save=False, local_execution=local_execution, **kwargs)

        # Set as run's parent
        run.add_relationship(Relationship.RUN_OF.value, self.key)
        run.save()

        # If execution is done by DHCore backend, return the object
        if not local_execution:
            if wait:
                return run.wait(log_info=log_info)
            return run

        # If local execution, build and launch run.
        # Detach the run from the main thread
        run.build()
        with ThreadPoolExecutor(max_workers=1) as executor:
            result = executor.submit(run.run)
            r = result.result()
        return r



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Any

from digitalhub.entities._base._base.entity import Base


class Extension(Base):
    """
    Set of extension metadata for an entity.
    """

    def __init__(
        self,
        name: str,
        id: str,
        kind: str,
        spec: dict[str, Any],
    ) -> None:
        self.name = name
        self.id = id
        self.kind = kind
        self.spec = spec

    @classmethod
    def from_dict(cls, obj: dict) -> Extension:
        """
        Return entity extension object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity extension.

        Returns
        -------
        Extension
            An entity extension object.
        """
        return cls(**obj)



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import importlib
from pathlib import Path
from types import ModuleType
from typing import Any

from digitalhub.utils.uri_utils import has_local_scheme, has_s3_scheme


def _load_croissant() -> ModuleType:
    """
    Load the mlcroissant module.

    Returns
    -------
    ModuleType
        The mlcroissant module.
    """
    try:
        return importlib.import_module("mlcroissant")
    except ImportError as e:
        raise ModuleNotFoundError("Please install 'mlcroissant' to use this feature.") from e


def get_croissant_dataset(metadata_json_path: str, mapping: dict | None = None) -> Any:
    """
    Get the Croissant Dataset object from the metadata JSON file.

    Parameters
    ----------
    metadata_json_path : str
        Path to the metadata JSON file.
    mapping : dict | None, optional
        Optional mapping to apply to the metadata, by default None.

    Returns
    -------
    mlcroissant.Dataset
        Croissant Dataset object.
    """
    mlcroissant = _load_croissant()
    return mlcroissant.Dataset(metadata_json_path, mapping=mapping)


def get_files_from_croissant(dataset: Any, metadata_path: str) -> list[str]:
    """
    Parse Croissant metadata to extract the local files described
    with FileObject structure, save the paths in a list of strings.

    Parameters
    ----------
    dataset : mlcroissant.Dataset
        The Croissant Dataset object.
    metadata_path : str
        Path to the metadata JSON file.

    Returns
    -------
    list[str]
        List of local file paths extracted from FileObject structures.
    """
    files: list[str] = []
    for file_object in dataset.metadata.file_objects:
        content_url = file_object.content_url
        if content_url is None or not has_local_scheme(content_url):
            continue
        abs_metadata_path = Path(metadata_path).resolve()
        abs_content_url = (abs_metadata_path.parent / Path(content_url)).resolve()
        pth = str(abs_content_url.relative_to(abs_metadata_path.parent))
        files.append(pth)
    return files


def get_metadata_fields_from_croissant(dataset: Any) -> tuple[str | None, str | None, list[str] | None]:
    """
    Extract name, description, and keywords from Croissant metadata.

    Parameters
    ----------
    dataset : mlcroissant.Dataset
        The Croissant Dataset object.

    Returns
    -------
    tuple[str | None, str | None, list[str] | None]
        Name, description, and keywords (labels) from metadata.
    """
    metadata = dataset.metadata
    name = getattr(metadata, "name", None)
    description = getattr(metadata, "description", None)
    keywords = getattr(metadata, "keywords", None)
    if keywords is not None and not isinstance(keywords, list):
        keywords = [str(keywords)]
    return name, description, keywords


def validate_croissant_file(src: str) -> None:
    """
    Validate if the provided file is a Croissant metadata JSON file
    named metadata.json.

    Parameters
    ----------
    src : str
        Path to the file to validate.
    """
    filepath = Path(src)
    if filepath.name != "metadata.json":
        raise ValueError("The provided file is not a valid Croissant metadata JSON file.")
    if not filepath.is_file():
        raise ValueError("The provided path does not point to a valid file.")
    try:
        _ = get_croissant_dataset(src)
    except Exception as e:
        raise ValueError("The provided file is not a valid Croissant metadata JSON file.") from e


def validate_output_path(path: str | None) -> None:
    """
    Validate the output path for the Croissant dataitem.

    Parameters
    ----------
    path : str | None
        Output path to validate.
    """
    if path is None:
        return
    if not has_s3_scheme(path):
        raise ValueError("The output path for Croissant dataitem must be an S3 path.")
    if not path.endswith("/"):
        raise ValueError("The output path for Croissant dataitem must be a partition path ending with '/'.")



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing
from abc import abstractmethod

from digitalhub.utils.exceptions import BuilderError

if typing.TYPE_CHECKING:
    from digitalhub.stores.readers.data._base.reader import DataframeReader
    from digitalhub.utils.types import Dataframe


class ReaderBuilder:
    ENGINE: str = None
    DATAFRAME_CLASS: Dataframe = None  # type: ignore

    def __init__(self):
        if self.ENGINE is None:
            raise BuilderError("ENGINE must be set.")
        if self.DATAFRAME_CLASS is None:
            raise BuilderError("DATAFRAME_CLASS must be set.")

    @abstractmethod
    def build(self, **kwargs) -> DataframeReader:
        """
        Build reader object.
        """

    @abstractmethod
    def map_dataframe_name(self, dataframe: Dataframe) -> str:  # type: ignore
        """
        Map dataframe type to engine name.
        """



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing
from pathlib import Path
from typing import Any

from digitalhub.entities._commons.utils import refresh_decorator
from digitalhub.entities.dataitem._base.entity import Dataitem
from digitalhub.entities.dataitem.croissant.utils import get_croissant_dataset, get_files_from_croissant
from digitalhub.stores.data.api import get_store
from digitalhub.utils.uri_utils import has_s3_scheme

if typing.TYPE_CHECKING:
    from digitalhub.entities.dataitem.croissant.spec import DataitemSpecCroissant
    from digitalhub.entities.dataitem.croissant.status import DataitemStatusCroissant


class DataitemCroissant(Dataitem):
    """
    DataitemCroissant class.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.spec: DataitemSpecCroissant
        self.status: DataitemStatusCroissant

    @refresh_decorator
    def as_dataset(self, overwrite: bool = False) -> Any:
        """
        Get the Croissant Dataset object from the Dataitem.

        Parameters
        ----------
        overwrite : bool
            Flag to indicate overwrite of local files.

        Returns
        -------
        mlcroissant.Dataset
            Croissant Dataset object.
        """
        metadata_json_path = self._get_metadata_json()
        dtst = get_croissant_dataset(metadata_json_path)
        self._collect_local_data(dtst, metadata_json_path, overwrite=overwrite)
        mapping = self._map_files_filepaths(dtst)
        return get_croissant_dataset(metadata_json_path, mapping=mapping)

    def _get_metadata_json(self) -> str:
        """
        Get the metadata JSON croissant file.

        Returns
        -------
        str
            Metadata file path.
        """
        if has_s3_scheme(self.spec.path):
            store = get_store(self.spec.path)
            return store.download(
                self.spec.path + "metadata.json",
                self._context().root / self.ENTITY_TYPE,
                overwrite=True,
            )
        return self.download(overwrite=True)

    def _collect_local_data(
        self,
        dataset: Any,
        metadata_path: str,
        overwrite: bool = False,
    ) -> None:
        """
        Collect local data files from Croissant metadata.

        Parameters
        ----------
        dataset : mlcroissant.Dataset
            Croissant Dataset object.
        metadata_path : str
            Path to the metadata JSON file.
        overwrite : bool
            Flag to indicate overwrite of local files.
        """
        if not has_s3_scheme(self.spec.path):
            return
        local_files = get_files_from_croissant(dataset, metadata_path)
        store = get_store(self.spec.path)
        for file_path in local_files:
            key = self.spec.path + str(Path(file_path))
            store.download(
                key,
                self._context().root / self.ENTITY_TYPE / Path(file_path),
                overwrite=overwrite,
            )

    def _map_files_filepaths(self, dataset: Any) -> dict | None:
        """
        Docstring per _map_files_filepaths

        Parameters
        ----------
        dataset : Any
            Croissant Dataset object.

        Returns
        -------
        dict | None
            Mapping of file paths.
        """
        if not has_s3_scheme(self.spec.path):
            return
        download_root = self._context().root / self.ENTITY_TYPE
        return {
            str(Path(file_path)): str((download_root / Path(file_path)).resolve())
            for file_path in get_files_from_croissant(dataset, self._get_metadata_json())
        }



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import typing
from pathlib import Path

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.dataitem._base.crud import log_base_dataitem
from digitalhub.entities.dataitem.croissant.utils import (
    get_croissant_dataset,
    get_files_from_croissant,
    get_metadata_fields_from_croissant,
    validate_croissant_file,
    validate_output_path,
)

if typing.TYPE_CHECKING:
    from digitalhub.entities.dataitem.croissant.entity import DataitemCroissant


def log_croissant(
    project: str,
    name: str,
    source: str,
    drop_existing: bool = False,
    path: str | None = None,
    **kwargs,
) -> DataitemCroissant:
    """
    Create and upload an object.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    source : str
        Metadata JSON file path.
    drop_existing : bool
        Whether to drop existing entity with the same name.
    path : str
        Destination path of the dataitem. If not provided, it's generated.
    **kwargs : dict
        New dataitem spec parameters.

    Returns
    -------
    DataitemCroissant
        Object instance.

    Examples
    --------
    >>> obj = log_croissant(project="my-project",
    >>>                     name="my-croissant",
    >>>                     source="./metadata.json")
    """
    validate_croissant_file(source)
    validate_output_path(path)

    # Get the dataset and files from the Croissant metadata
    dataset = get_croissant_dataset(source)
    files = get_files_from_croissant(dataset, source)

    # Change working dir to the metadata.json parent to
    # correctly resolve relative paths in content_url
    current_dir = os.getcwd()
    Path(source).parent.resolve()
    os.chdir(Path(source).parent)
    sources = files + ["metadata.json"]
    kwargs["keep_dir_structure"] = True

    # Log dataitem in the context of the metadata.json dir
    dataitem = log_base_dataitem(
        project=project,
        name=name,
        kind=EntityKinds.DATAITEM_CROISSANT.value,
        source=sources,
        drop_existing=drop_existing,
        path=path,
        **kwargs,
    )
    os.chdir(current_dir)

    # Update metadata fields from Croissant
    metadata = get_metadata_fields_from_croissant(dataset)
    dataitem.metadata.name = metadata[0]
    dataitem.metadata.description = metadata[1]
    if metadata[2] is not None:
        dataitem.metadata.labels = metadata[2]
    dataitem.save(update=True)

    return dataitem



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.context.context import Context
from digitalhub.utils.exceptions import ContextError

if typing.TYPE_CHECKING:
    from digitalhub.entities.project._base.entity import Project


class ContextBuilder:
    """
    A builder class for managing project contexts.
    """

    def __init__(self) -> None:
        self._instances: dict[str, Context] = {}

    def build(self, project: Project) -> Context:
        """
        Add a project as context and return the created Context instance.

        Parameters
        ----------
        project : Project
            The project instance to create a context for.

        Returns
        -------
        Context
            The newly created or existing Context instance.
        """
        if project.name not in self._instances:
            # Create context without registering it yet to avoid recursion
            # then register it before initialization to prevent recursion
            # finally initialize it
            ctx = Context.__new__(Context)
            self._instances[project.name] = ctx
            ctx.__init__(project)
        return self._instances[project.name]

    def get(self, project: str) -> Context:
        """
        Retrieve a context instance by project name.

        Parameters
        ----------
        project : str
            The name of the project whose context to retrieve.

        Returns
        -------
        Context
            The context instance associated with the project.
        """
        try:
            return self._instances[project]
        except KeyError:
            raise ContextError(f"Context '{project}' not found. Get or create a project named '{project}'.")

    def remove(self, project: str) -> None:
        """
        Remove a project's context from the registry.

        Parameters
        ----------
        project : str
            The name of the project whose context should be removed.
        """
        self._instances.pop(project, None)


context_builder = ContextBuilder()
